/*  Copyright (C) 1996-2001  Martin Vicente			      -*- c -*-
    This file is part of the eXtended C Library (XCLIB).

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library (see the file COPYING.LIB).
    If not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.		     */


/****  Project : eXtended C Library (XCLIB)
    *  Purpose : GENERIC LIST
    *  Language: ANSI C
    *  File    : demo_list-6.c
    *  Author  : Martin Vicente <martin@vicente.org>
    *
    ****/


/*****************************************************************************\

 [English]

 <xc/list> use with type checking.

 Demonstration of the hard disk writing of an object list, and its reading,
 in user mode. This mode allows to create a database in ASCII format;
 which is therefore readable and completely portable from a system to
 another one.

 [French]

 Utilisation de <xc/list> avec contrle de type.

 Dmonstration de l'criture sur le disque d'une liste d'objets, et de sa
 lecture, en mode utilisateur. Ce mode permet l'tablissement d'une base de
 donnes au format ASCII; donc lisible et parfaitement portable d'un systme
  l'autre.

\*****************************************************************************/


#include <stddef.h>
#include <stdio.h>
#include <assert.h>

#include <xc/strlib.h>
#include <xc/error.h>
#include <xc/xalloc.h>
#include <xc/list.h>

typedef struct ObjectRec *Object;

def_LIST(OBJECTS,, Object);

#define OBJECTS	(*OBJECTS)	/* optional */

#define FILENAME	"objects.dat"

struct ObjectRec
{
    int   number;
    char *name;
};

static Object createObject(const char *name);
static void   destroyObject(Object object);

static int print_object(ObjectId id, void *data);
static int destroy_object(ObjectId id, void *data);
static int write_object(ObjectId id, void *data, FILE *file);
static int read_object(ObjectId id, void *data, FILE *file);

#define COMM(en, fr)	__xcint(comm)(en, fr, NULL)
#define WAIT()		__xcint(wait)()

extern void __xcint(comm)(const char *en, ...);	/* DON'T USE IT! */
extern void __xcint(wait)(void);		/* DON'T USE IT! */

int main()
{
    ObjectList list;

    COMM("Creating list of objects",
	 "Cration de la liste d'objets");

    list = OBJECTS.create(sizeof (struct ObjectRec));

    OBJECTS.add(list, createObject("Black"));
    OBJECTS.add(list, createObject("Red"));
    OBJECTS.add(list, createObject("Green"));
    OBJECTS.add(list, createObject("Yellow"));
    OBJECTS.add(list, createObject("Blue"));
    OBJECTS.add(list, createObject("Magenta"));
    OBJECTS.add(list, createObject("Cyan"));
    OBJECTS.add(list, createObject("White"));

    OBJECTS.action(list, print_object, NULL);

    WAIT();

    COMM("Writing to disk",
	 "Ecriture sur disque de la liste d'objets");

    {
	FILE *file;

	if ((file = fopen(FILENAME, "w")) == NULL)
	    error("Can't open file");

	if (OBJECTS.write(list, write_object, NULL, file) == EOF)
	    error("Can't write file");

	if (fclose(file) == EOF)
	    error("Can't close file");
    }

    COMM("Destroying object's list",
	 "Destruction de la liste d'objets");

    OBJECTS.action(list, destroy_object, NULL);
    OBJECTS.destroy(list);

    WAIT();

    COMM("Reading list from disk",
	 "Cration de la liste d'objets depuis le fichier disque");

    {
	FILE *file;

	if ((file = fopen(FILENAME, "r")) == NULL)
	    error("Can't open file");

	if ((list = OBJECTS.read(read_object, NULL, file)) == NULL)
	    error("Can't read file");

	if (fclose(file) == EOF)
	    error("Can't close file");
    }

    OBJECTS.action(list, print_object, NULL);

    WAIT();

    COMM("Destroying object's list",
	 "Destruction de la liste d'objets");

    OBJECTS.action(list, destroy_object, NULL);
    OBJECTS.destroy(list);

    return 0;
}

static Object createObject(const char *name)
{
    static int number = 0;

    Object object;

    assert(name != NULL);

    object = xnew(Object);

    object->number = ++number;
    object->name   = xstrdup(name);

    return object;
}

static void destroyObject(Object object)
{
    assert(object != NULL);

    xfree(object->name);

    xfree(object);
}

static int print_object(ObjectId id, __unused void *data)
{
    printf("\tObject %d - %s\n", id->item->number, id->item->name);

    return 0;
}

static int destroy_object(ObjectId id, __unused void *data)
{
    destroyObject(id->item);

    return 0;
}

static int write_object(ObjectId id, __unused void *data, FILE *file)
{
    if (fprintf(file, "%d\t%s\n", id->item->number, id->item->name) < 0)
	return EOF;

    return 0;
}

static int read_object(ObjectId id, __unused void *data, FILE *file)
{
    int  number;
    char name[8];

    if (fscanf(file, "%d\t%s\n", &number, name) == EOF)
	return EOF;

    OBJECTS.alloc(id, sizeof (struct ObjectRec));

    id->item->number = number;
    id->item->name   = xstrdup(name);

    return 0;
}


/* End of file */
