/*  XCLIB - a set of C functions for general purposes		      -*- c -*-
    Copyright (C) 1996-2001  Martin Vicente

    This file is part of the eXtended C Library (XCLIB).

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library (see the file COPYING.LIB).
    If not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.		     */


/*|+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++|*\
  |  Martin Vicente, 1996-2001
  |
  | Project : XCLIB (eXtended C Library)
  | Purpose : PARSE COMMAND LINE OPTIONS
  | Version : 0.3
  | Language: ANSI C
  | File    : options/options.c
  | Author  : Martin VICENTE <martin@vicente.org>
\*|+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++|*/


/*****************************************************************************\

Il me faut finir de traiter les cas d'erreur dans les diffrentes conversions.

\*****************************************************************************/


#define _XCLIB_INTERNALS

#define _FACILITY	FACILITY(option)

#include <stddef.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include <xc/bool.h>
#include <xc/errors.h>
#include <xc/program.h>
#include <xc/error.h>

#include "xc/options.h"
#
#undef parse_options

#include "global.h"

#define ERR_AMBIGUOUS	"Ambiguous option `%s'"
#define ERR_TRYHELP	"Try `%s --help' for more information"

#define OK		0
#define NOT_FOUND	1

#define EQ(s1, s2)	(strcmp(s1, s2) == 0)

static void reset_options(void *base, const Option *options, size_t noptions);
static void parse_options(void *base, const Option *options, size_t noptions, int *argc, char **argv);
static void error_options(const Option *options, size_t noptions, const char *format, ...) __noreturn;

void __xclib(parse_options)(
    void         *base,
    Option const *options,
    size_t        noptions,
    int          *argc,
    char        **argv)
{
    assert(base != NULL && options != NULL && argc != NULL && argv != NULL);

    reset_options(base, options, noptions);

    parse_options(base, options, noptions, argc, argv);
}

static void reset(void *base, const Option *option);
static void setup(void *base, const Option *option, int *argc, char **argv);

static void reset_options(void *base, const Option *options, size_t noptions)
{
    register size_t i;

    for (i = 0; i < noptions; ++i) reset(base, &options[i]);
}

static bool check(void *base, const Option *options, size_t noptions, int *argc, char **argv);

static void compress_args(int *argc, char **argv);

static void parse_options(void *base, const Option *options, size_t noptions, int *argc, char **argv)
{
    ++argv;

    while (*argv != NULL)
    {
	if (EQ(*argv, "-"))
	{
	    ++argv;
	}
	else if (EQ(*argv, "--"))
	{
	    compress_args(argc, argv);

	    break;
	}
	else if ((*argv)[0] == '-')
	{
	    if (check(base, options, noptions, argc, argv) == NOT_FOUND)
	    {
		++argv;
	    }
	}
	else
	{
	    ++argv;
	}
    }
}

static bool check(void *base, const Option *options, size_t noptions, int *argc, char **argv)
{
    size_t len, n;

    register size_t i;

    len = strlen(*argv);

    n = 0;

    for (i = 0; i < noptions; ++i)
    {
	if (strncmp(*argv, options[i].name, len) == 0)
	{
	    if (++n > 1)
	    {
		error_options(options, noptions, ERR_AMBIGUOUS, argv[0]);
	    }
	}
    }

    for (i = 0; i < noptions; ++i)
    {
	if (strncmp(*argv, options[i].name, len) == 0)
	{
	    setup(base, &options[i], argc, argv);

	    break;
	}
    }

    if (i < noptions) return OK;

    error_options(options, noptions, ERR_INVOPT" `%s'", argv[0]);

    return NOT_FOUND;
}

static void error_options(const Option *options, size_t noptions, const char *format, ...)
{
    register size_t i;

    clrerr();

    for (i = 0; i < noptions; ++i)
    {
	if (EQ("--help", options[i].name))
	{
	    seterr1(ERR_TRYHELP".", program_invocation_short_name);

	    break;
	}
    }

    {
	va_list vargs;

	va_start(vargs, format);
	vputerr("Error", format, vargs);
	va_end(vargs);
    }

    exit(EXIT_INVALID);
}

static void set(void *base, const Option *option, OptionValue value);

static void reset(void *base, const Option *option)
{
    set(base, option, option->reset_value);
}

static void setup(void *base, const Option *option, int *argc, char **argv)
{
    switch (option->kind)
    {
    case OptionKindNoArg:
    {
	compress_args(argc, argv);

	set(base, option, option->setup_value);
    }
    break;

    case OptionKindSepArg:
    {
	switch (option->type)
	{
	case OptionTypeChar:
	{
	    char value;

	    compress_args(argc, argv);

	    if (*argv == NULL) error1(ERR_NOOPTARG, option->name);

	    if ((*argv)[0] == '\0' || (*argv)[1] != '\0') error1(ERR_INVOPTARG" `%s'", *argv);

	    value = (*argv)[0];

#ifdef __GNUC__
	    set(base, option, (OptionValue)value);
#else
	    fatal(ERR_PROGRAM);
#endif

	    compress_args(argc, argv);
	}
	break;

	case OptionTypeBool:
	{
	    bool value;

	    compress_args(argc, argv);

	    if (*argv == NULL) error1(ERR_NOOPTARG, option->name);

	    value = atoi(*argv);

#ifdef __GNUC__
	    set(base, option, (OptionValue)value);
#else
	    fatal(ERR_PROGRAM);
#endif

	    compress_args(argc, argv);
	}
	break;
	    
	case OptionTypeInt:
	{
	    char *end;
	    int   value;

	    compress_args(argc, argv);

	    if (*argv == NULL) error1(ERR_NOOPTARG, option->name);

	    value = strtol(*argv, &end, 0);

	    if (end != NULL && end[0] != '\0') exit(1);

#ifdef __GNUC__
	    set(base, option, (OptionValue)value);
#else
	    fatal(ERR_PROGRAM);
#endif

	    compress_args(argc, argv);
	}
	break;

	case OptionTypeFloat:
	{
	    char *end;
	    float value;

	    compress_args(argc, argv);

	    if (*argv == NULL) error1(ERR_NOOPTARG, option->name);

	    value = strtod(*argv, &end);

	    if (end != NULL && end[0] != '\0')
		error2(ERR_INVOPTARG" `%s' of option `%s'", *argv, option->name);

#ifdef __GNUC__
	    set(base, option, (OptionValue)value);
#else
	    fatal(ERR_PROGRAM);
#endif

	    compress_args(argc, argv);
	}
	break;

	case OptionTypeDouble:
	{
	    char  *end;
	    double value;

	    compress_args(argc, argv);

	    if (*argv == NULL) error1(ERR_NOOPTARG, option->name);

	    value = strtod(*argv, &end);

	    if (end != NULL && end[0] != '\0')
		error2(ERR_INVOPTARG" `%s' of option `%s'", *argv, option->name);

#ifdef __GNUC__
	    set(base, option, (OptionValue)value);
#else
	    fatal(ERR_PROGRAM);
#endif

	    compress_args(argc, argv);
	}
	break;

	case OptionTypeString:
	{
	    compress_args(argc, argv);

	    if (*argv == NULL) error1(ERR_NOOPTARG, option->name);

#ifdef __GNUC__
	    set(base, option, (OptionValue)(void *)*argv);
#else
	    fatal(ERR_PROGRAM);
#endif

	    compress_args(argc, argv);
	}
	break;

	case OptionTypePointer:
	{
	    void *value;

	    compress_args(argc, argv);

	    if (*argv == NULL) error1(ERR_NOOPTARG, option->name);

	    if (sscanf(*argv, "%p", &value) < 1)
		error2(ERR_INVOPTARG" `%s' of option `%s'", *argv, option->name);

#ifdef __GNUC__
	    set(base, option, (OptionValue)value);
#else
	    fatal(ERR_PROGRAM);
#endif

	    compress_args(argc, argv);
	}
	break;
	}
    }
    break;
    }
}

static void compress_args(int *argc, char **argv)
{
    char **arg;

    assert(argc != NULL && argv != NULL);

    --*argc;

    for (arg = argv; *arg != NULL; ++arg) *arg = *(arg + 1);
}

static void set(void *base, const Option *option, OptionValue value)
{
    switch (option->type)
    {
    case OptionTypeChar:    *((char        *)((char *)base + option->offset)) = value._char;    break;
    case OptionTypeBool:    *((bool        *)((char *)base + option->offset)) = value._bool;    break;
    case OptionTypeInt:     *((int         *)((char *)base + option->offset)) = value._int;     break;
    case OptionTypeFloat:   *((float       *)((char *)base + option->offset)) = value._float;   break;
    case OptionTypeDouble:  *((double      *)((char *)base + option->offset)) = value._double;  break;
    case OptionTypeString:  *((const char **)((char *)base + option->offset)) = value._string;  break;
    case OptionTypePointer: *((void       **)((char *)base + option->offset)) = value._pointer; break;
    default:
	fatal(ERR_INTERNAL);
	break;
    }
}


/* End Of File */
